<?php

require __DIR__ . '/SolarController.php';
require __DIR__ . '/SolarStation.php';
require __DIR__ . '/Logger.php';

use Nette\Http\UrlScript;
use Nette\Routing\Router;
use Nette\SmartObject;

class Application
{
    use SmartObject;

    private Router $router;
    private Logger $logger;
    private \React\Http\HttpServer $httpServer;
    private SolarController $controller;

    public function bootstrap(): self
    {
        $this->createRouter();
        $this->logger = new Logger();
        $this->setpControler();
        $this->setupHttpServer();

        return $this;
    }

    public function run(string $host)
    {
        $socket = new React\Socket\SocketServer($host);
        $this->httpServer->listen($socket);

        $loop = React\EventLoop\Loop::get();
        $loop->addPeriodicTimer(3, fn () => $this->controller->refresh());

        $this->logger->log("SERVER", "running at $host");
    }

    private function createRouter()
    {
        $this->router = new \Nette\Routing\RouteList();
        $this->router->addRoute('/start/<outNum>', ['action' => 'start']);
        $this->router->addRoute('/status', ['action' => 'status']);
        $this->router->addRoute('/setStatus/<outNum>/<requiredStatus>[/<duration>]', ['action' => 'setStatus']);
        $this->router->addRoute('/testing', ['action' => 'testing']);
        $this->router->addRoute('/dist/<path(.+)>', ['action' => 'dist']);
    }

    private function setpControler()
    {
        $this->controller = new SolarController(
            $_ENV['CONTROLLER_URL'],
            $this->logger
        );
        $this->controller->addStation(
            new SolarStation(0)
        );

    }

    private function setupHttpServer()
    {
        $this->httpServer = new React\Http\HttpServer(function (Psr\Http\Message\ServerRequestInterface $request) {
            try {
                $res = $this->processRequest($request);
                if (is_array($res)) $hres = React\Http\Message\Response::json($res);
                else if (is_string($res)) $hres = React\Http\Message\Response::plaintext($res);
                else if ($res instanceof React\Http\Message\Response) $hres = $res;
                $hres = $this->setupCors($hres, $request);

                return $hres;
            } catch (Exception $e) {
                $this->logger->log("ERROR", $e);
            }
        });
        $this->httpServer->on('error', fn($error) => $this->logger->log("ERROR", $error));
    }

    private function processRequest(\Psr\Http\Message\ServerRequestInterface $request)
    {
        $url = $request->getUri()->__toString();
        $this->logger->log('URL', $url);
        $nRequest = new \Nette\Http\Request(new UrlScript($url));
        $res = $this->router->match($nRequest);
        if (!$res) return new React\Http\Message\Response(404, [], "Not found " . $url);
        $methodName = 'action' . \Nette\Utils\Strings::firstUpper($res['action']);
        $params = $res;
        unset($res['action']);
        return $this->$methodName(...array_values($params));
        // $this->logger->log($res);
    }

    private function setupCors(React\Http\Message\Response $response, Psr\Http\Message\ServerRequestInterface $request): React\Http\Message\Response
    {
        $headers = array_map(
            fn($v) => $v[0],
            $request->getHeaders()
        );
        $origin = isset($headers['Origin']) ? $headers['Origin'] : (isset($headers['Referer']) ? $headers['Referer'] : '*');
        $response = $response
            ->withHeader("Access-Control-Allow-Origin", $origin)
            ->withHeader("Access-Control-Allow-Credentials", "true")
            ->withHeader("Access-Control-Max-Age", "1"/*"86400"*/)
            ->withHeader("Access-Control-Allow-Methods", "GET, POST, PUT, DELETE, OPTIONS, PATCH, HEAD")
            ->withHeader('Access-Control-Allow-Headers', 'X-Requested-With, Content-Type, Accept, Origin, Authorization, Cookie');

        if ($request->getMethod() == 'OPTIONS') {

            if (isset($headers['Access-Control-Request-Methods']))
                $response = $response->withHeader("Access-Control-Allow-Methods", "GET, POST, OPTIONS");

            if (isset($headers['Access-Control-Request-Headers']))
                $response = $response->withHeader("Access-Control-Allow-Headers", $headers['Access-Control-Request-Headers']);
        }

        return $response;
    }

    public function actionStart($outNum) {
        return $this->controller->start((int)$outNum);
    }

    public function actionStatus(): array {
        return $this->controller->status();
    }

    public function actionSetStatus($outNum, $requiredStatus, $duration)
    {
        $res = $this->controller->setStatus($outNum, $requiredStatus, $duration ? (int)$duration : null);
        if (is_string($res)) return "ERROR: " . $res;
        return $this->controller->status();
    }

}